#ident	"@(#)slip.c	1.2"
/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
/*
 * SLIP - Serial Line IP module/driver (ala RFC1055)
 */
#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/errno.h>
#include <sys/kmem.h>
#include <sys/cred.h>
#include <sys/dlpi.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <net/if.h>
#include <netinet/in.h>
#include <sys/cmn_err.h>
#include <sys/ddi.h>
#include <slip.h>

#undef STATIC
#ifdef DEBUG
#define STATIC
#else
#define STATIC static
#endif

#define HALFSEC		500000	/* in microseconds */

/* interface flags */
#define IFF_SLIPFLAGS	(IFF_POINTOPOINT|IFF_RUNNING)

typedef struct slip slip_t;

struct slip {
	queue_t		*s_rdq;		/* read queue */
	slip_t		*s_next;	/* linked list */
	ulong_t		s_sap;		/* sap bound */
	ulong_t		s_rid;		/* read bufcall and timeout ID */
	ulong_t		s_wid;		/* write bufcall and timeout ID */
	ulong_t		s_flag;		/* see below */
	mblk_t		*s_udmp;	/* unitdata indication */
					/* being assembled */
	mblk_t		*s_hupmp;	/* used to notify daemon of */
					/* hangup */
	dev_t		s_dev;		/* device number of TTY driver */
	ushort_t	s_dlstate;	/* DLPI state */
	struct ifstats	s_ifstats;	/* for netstat(1) */
	ulong_t		s_ifflag;	/* interface flags */
	struct ifaddr	s_ifaddr;	/* interface addresses */
	char		s_ifname[IFNAMSIZ];/* interface name */
};

/* flags */
#define RDBCALL	0x01	/* read side bufcall pending */
#define RDTOUT	0x02	/* read side timeout pending */
#define WRBCALL	0x04	/* write side bufcall pending */
#define WRTOUT	0x08	/* write side timeout pending */
#define GOTESC	0x10	/* received an escape */
#define LOCADDR	0x20	/* local address set */
#define REMADDR	0x40	/* remote address set */

STATIC slip_t slipadmin;
STATIC slip_t *sliplist;

int slipdevflag = 0;

extern struct ifstats *ifstats;	/* global kernel list */

typedef enum retval {
	DONE,		/* function completed its job */
	RETRY,		/* temporary failure; try again later */
	ERR		/* permanent error */
} retval_t;

STATIC int slipopen(queue_t *, dev_t *, int, int, cred_t *);
STATIC int slipclose(queue_t *, int, cred_t *);
STATIC int slipwput(queue_t *, mblk_t *);
STATIC int slipwsrv(queue_t *);
STATIC int sliprput(queue_t *, mblk_t *);
STATIC int sliprsrv(queue_t *);
STATIC void slip_ioctl(queue_t *, mblk_t *);
STATIC void slip_hangup(queue_t *);
STATIC retval_t slip_info(slip_t *, mblk_t *);
STATIC retval_t slip_bind(slip_t *, mblk_t *);
STATIC retval_t slip_unbind(slip_t *, mblk_t *);
STATIC retval_t slip_send(slip_t *, mblk_t *);
STATIC retval_t slip_errorack(slip_t *, mblk_t *, ulong_t);
STATIC void slip_wsched(slip_t *, int);
STATIC void slip_wcont(slip_t *);
STATIC void slip_rsched(slip_t *, int);
STATIC void slip_rcont(slip_t *);

STATIC struct module_info slip_minfo = {
	0x534c, "slip", MINSDUSZ, MAXSDUSZ, 8*MAXSDUSZ, 2*MAXSDUSZ
};
STATIC struct qinit slip_rinit = {
	sliprput, sliprsrv, slipopen, slipclose, NULL,
	&slip_minfo, NULL
};
STATIC struct qinit slip_winit = {
	slipwput, slipwsrv, NULL, NULL, NULL, &slip_minfo, NULL
};
struct streamtab slipinfo = {
	&slip_rinit, &slip_winit, NULL, NULL
};

STATIC int
slipopen(queue_t *q, dev_t *devp, int flag, int sflag,
	cred_t *crp)
{
	slip_t *slp;
	mblk_t *bp;

	switch (sflag) {
	default:
		return(EINVAL);

	case MODOPEN:
		if (q->q_ptr != NULL)	/* not first open */
			return(0);
		bp = allocb(sizeof(dev_t), BPRI_HI);
		if (bp == NULL)
			return(EAGAIN);
		slp = (slip_t *)kmem_zalloc(sizeof(slip_t), KM_SLEEP);
		slp->s_hupmp = bp;
		slp->s_next = sliplist;
		sliplist = slp;
		break;

	case 0:
		if (geteminor(*devp) == ADMINDEV) {
			slp = &slipadmin;
			if (slp->s_rdq != NULL)
				return(EBUSY);	/* one at a time */
		} else {
			return(ENXIO);
		}
		slp->s_hupmp = NULL;
		break;
	}
	slp->s_rdq = q;
	slp->s_dlstate = DL_UNBOUND;
	slp->s_flag = 0;
	slp->s_ifflag = IFF_POINTOPOINT;
	slp->s_udmp = NULL;
	slp->s_dev = *devp;
	slp->s_ifstats.ifs_name = slp->s_ifname;
	slp->s_ifstats.ifs_mtu = MAXSDUSZ;
	slp->s_ifstats.ifs_addrs = &slp->s_ifaddr;
	slp->s_ifaddr.ifa_ifs = &slp->s_ifstats;
	q->q_ptr = (caddr_t)slp;
	WR(q)->q_ptr = (caddr_t)slp;
	return(0);
}

STATIC int
slipclose(queue_t *q, int flag, cred_t *crp)
{
	slip_t *slp, *tslp, *pslp;
	struct ifstats *ifp, *pifp;

	slp = (slip_t *)q->q_ptr;
	if (slp->s_flag & RDTOUT)
		untimeout(slp->s_rid);
	else if (slp->s_flag & RDBCALL)
		unbufcall(slp->s_rid);
	if (slp->s_flag & WRTOUT)
		untimeout(slp->s_wid);
	else if (slp->s_flag & WRBCALL)
		unbufcall(slp->s_wid);
	if (slp != &slipadmin)
		slip_hangup(q);
	pifp = NULL;
	ifp = ifstats;
	while ((ifp != NULL) && (ifp != &slp->s_ifstats)) {
		pifp = ifp;
		ifp = ifp->ifs_next;
	}
	if (ifp != NULL) {
		if (pifp != NULL)
			pifp->ifs_next = ifp->ifs_next;
		else
			ifstats = ifp->ifs_next;
	}
	tslp = sliplist;
	pslp = NULL;
	while ((tslp != NULL) && (tslp != slp)) {
		pslp = tslp;
		tslp = tslp->s_next;
	}
	if (tslp == slp) {
		if (pslp != NULL)
			pslp->s_next = slp->s_next;
		else
			sliplist = slp->s_next;
		kmem_free(slp, sizeof(slip_t));
	} else {
		slp->s_rdq = NULL;
	}
	q->q_ptr = NULL;
	WR(q)->q_ptr = NULL;
	return(0);
}

STATIC int
slipwput(queue_t *q, mblk_t *mp)
{
	slip_t *slp;

	slp = (slip_t *)q->q_ptr;
	if (slp == &slipadmin) {
		if (mp->b_datap->db_type == M_IOCTL) {
			mp->b_datap->db_type = M_IOCNAK;
			qreply(q, mp);
		} else {
			freemsg(mp);
		}
		return(0);
	}
	switch (mp->b_datap->db_type) {
	case M_FLUSH:
		if (*mp->b_rptr & FLUSHW)
			flushq(q, FLUSHALL);
		if (*mp->b_rptr & FLUSHR)
			flushq(RD(q), FLUSHALL);
		putnext(q, mp);
		break;

	case M_DATA:
		freemsg(mp);
		break;

	case M_PROTO:
	case M_PCPROTO:
		putq(q, mp);
		break;

	case M_IOCTL:
		slip_ioctl(q, mp);
		break;

	default:
		putnext(q, mp);
		break;
	}
	return(0);
}

STATIC void
slip_ioctl(queue_t *q, mblk_t *mp)
{
	slip_t *slp;
	struct iocblk *iocp;
	struct iocblk_in *iniocp;
	struct ifreq *reqp;
	int i;
	ushort_t unit;

	slp = (slip_t *)q->q_ptr;
	iocp = (struct iocblk *)mp->b_rptr;
	switch (iocp->ioc_cmd) {
	case SIOCGIFFLAGS:
		if ((mp->b_cont == NULL) ||
		  (iocp->ioc_count == TRANSPARENT)) {
			mp->b_datap->db_type = M_IOCNAK;
		} else {
			reqp = (struct ifreq *)mp->b_cont->b_rptr;
			reqp->ifr_flags = slp->s_ifflag;
			mp->b_datap->db_type = M_IOCACK;
		}
		break;

	case SIOCSIFFLAGS:
		if ((mp->b_cont == NULL) ||
		  (iocp->ioc_count == TRANSPARENT)) {
			mp->b_datap->db_type = M_IOCNAK;
		} else {
			reqp = (struct ifreq *)mp->b_cont->b_rptr;
			slp->s_ifflag = reqp->ifr_flags |
			  (slp->s_ifflag & IFF_SLIPFLAGS);
			mp->b_datap->db_type = M_IOCACK;
		}
		break;

	case SIOCSIFNAME:
		if ((mp->b_cont == NULL) ||
		  (iocp->ioc_count == TRANSPARENT) ||
		  (slp->s_ifname[0] != '\0')) {
			mp->b_datap->db_type = M_IOCNAK;
		} else {
			reqp = (struct ifreq *)mp->b_cont->b_rptr;
			for (i = 0; i < IFNAMSIZ; i++)
				if ((reqp->ifr_name[i] == '\0') ||
				  ((reqp->ifr_name[i] >= '0') &&
				  (reqp->ifr_name[i] <= '9')))
					break;
				else
					slp->s_ifname[i] = reqp->ifr_name[i];
			for (unit = 0; i < IFNAMSIZ; i++)
				if (reqp->ifr_name[i] == '\0')
					break;
				else
					unit = unit * 10 + reqp->ifr_name[i] - '0';
			slp->s_ifstats.ifs_unit = unit;
			mp->b_datap->db_type = M_IOCACK;
		}
		break;

	case SIOCSIFADDR:
		if ((slp->s_flag & LOCADDR) || (mp->b_cont == NULL) ||
		  (iocp->ioc_count == TRANSPARENT)) {
			mp->b_datap->db_type = M_IOCNAK;
		} else {
			reqp = (struct ifreq *)mp->b_cont->b_rptr;
			slp->s_ifaddr.ifa_addr = reqp->ifr_addr;
			slp->s_flag |= LOCADDR;
			mp->b_datap->db_type = M_IOCACK;
			if ((slp->s_flag & (LOCADDR|REMADDR)) ==
			  (LOCADDR|REMADDR)) {
				slp->s_ifstats.ifs_next = ifstats;
				ifstats = &slp->s_ifstats;
				slp->s_ifflag |= IFF_RUNNING;
				slp->s_ifstats.ifs_active = 1;
			}
		}
		break;

	case SIOCSIFDSTADDR:
		if ((slp->s_flag & REMADDR) || (mp->b_cont == NULL) ||
		  (iocp->ioc_count == TRANSPARENT)) {
			mp->b_datap->db_type = M_IOCNAK;
		} else {
			reqp = (struct ifreq *)mp->b_cont->b_rptr;
			slp->s_ifaddr.ifa_dstaddr = reqp->ifr_dstaddr;
			slp->s_flag |= REMADDR;
			mp->b_datap->db_type = M_IOCACK;
			if ((slp->s_flag & (LOCADDR|REMADDR)) ==
			  (LOCADDR|REMADDR)) {
				slp->s_ifstats.ifs_next = ifstats;
				ifstats = &slp->s_ifstats;
				slp->s_ifflag |= IFF_RUNNING;
				slp->s_ifstats.ifs_active = 1;
			}
		}
		break;

	case SIOCGIFNETMASK:
	case SIOCSIFNETMASK:
	case SIOCGIFADDR:
	case SIOCGIFDSTADDR:
	case SIOCGIFMETRIC:
		mp->b_datap->db_type = M_IOCACK;
		break;

	default:
		putnext(q, mp);
		return;
	}
	if ((mp->b_datap->db_type == M_IOCACK) &&
	  ((mp->b_wptr - mp->b_rptr) >= sizeof(struct iocblk_in))) {
		iniocp = (struct iocblk_in *)mp->b_rptr;
		iniocp->ioc_ifflags = slp->s_ifflag;
	}
	qreply(q, mp);
}

STATIC int
slipwsrv(queue_t *q)
{
	mblk_t *mp;
	slip_t *slp;
	union DL_primitives *dlp;
	retval_t ret;

	if ((slp = (slip_t *)q->q_ptr) == NULL)
		return(0);
	while ((mp = getq(q)) != NULL) {
		dlp = (union DL_primitives *)mp->b_rptr;
		switch (dlp->dl_primitive) {
		case DL_UNITDATA_REQ:
			ret = slip_send(slp, mp);
			break;

		case DL_INFO_REQ:
			ret = slip_info(slp, mp);
			break;

		case DL_BIND_REQ:
			ret = slip_bind(slp, mp);
			break;

		case DL_UNBIND_REQ:
			ret = slip_unbind(slp, mp);
			break;

		default:
			ret = slip_errorack(slp, mp, DL_BADPRIM);
			break;
		}
		if (ret == RETRY) {
			putbq(q, mp);
			break;
		}
	}	/* while */
	return(0);
}

STATIC retval_t
slip_info(slip_t *slp, mblk_t *mp)
{
	dl_info_ack_t *ackp;
	mblk_t *bp;

	bp = allocb(DL_INFO_ACK_SIZE+SLIPSAPSZ, BPRI_HI);
	if (bp == NULL) {
		cmn_err(CE_WARN, "slip: can't acknowledge DL_INFO_REQ");
		return(ERR);
	}
	freemsg(mp);
	bp->b_datap->db_type = M_PCPROTO;
	ackp = (dl_info_ack_t *)bp->b_wptr;
	ackp->dl_primitive = DL_INFO_ACK;
	ackp->dl_max_sdu = MAXSDUSZ;
	ackp->dl_min_sdu = MINSDUSZ;
	ackp->dl_addr_length = sizeof(ulong_t);
#ifdef DLPIV2
	ackp->dl_mac_type = DL_OTHER;
#else
	ackp->dl_mac_type = DL_ETHER;
#endif
	ackp->dl_reserved = 0;
	ackp->dl_current_state = slp->s_dlstate;
#ifdef DLPIV2
	ackp->dl_sap_length = sizeof(ulong_t);
#else
	ackp->dl_reserved2 = 0;
#endif
	ackp->dl_service_mode = DL_CLDLS;
	ackp->dl_qos_length = 0;
	ackp->dl_qos_offset = 0;
	ackp->dl_qos_range_length = 0;
	ackp->dl_qos_range_offset = 0;
	ackp->dl_provider_style = DL_STYLE1;
	bp->b_wptr += DL_INFO_ACK_SIZE;
	if (slp->s_dlstate == DL_IDLE) {
		ackp->dl_addr_offset = DL_INFO_ACK_SIZE;
		*(long *)bp->b_wptr = slp->s_sap;
		bp->b_wptr += SLIPSAPSZ;
	} else {
		ackp->dl_addr_offset = 0;
	}
#ifdef DLPIV2
	ackp->dl_version = DL_VERSION_2;
	ackp->dl_brdcst_addr_length = 0;
	ackp->dl_brdcst_addr_offset = 0;
#endif
	ackp->dl_growth = 0;
	putnext(slp->s_rdq, bp);
	return(DONE);
}

STATIC void
slip_wsched(slip_t *slp, int size)
{
	ulong_t id;

	if (slp->s_flag & (WRBCALL|WRTOUT))
		return;
	id = bufcall(size, BPRI_HI, slip_wcont, slp);
	if (id == 0) {
		id = timeout(slip_wcont, (caddr_t)slp,
		  drv_usectohz(HALFSEC));
		slp->s_flag |= WRTOUT;
	} else {
		slp->s_flag |= WRBCALL;
	}
	slp->s_wid = id;
}

STATIC void
slip_wcont(slip_t *slp)
{
	qenable(WR(slp->s_rdq));
	slp->s_flag &= ~(WRBCALL|WRTOUT);
}

STATIC retval_t
slip_bind(slip_t *slp, mblk_t *mp)
{
	slip_t *tslp;
	dl_bind_req_t *reqp;
	dl_bind_ack_t *ackp;
	mblk_t *bp;

	if (slp->s_dlstate != DL_UNBOUND) {
		if (slip_errorack(slp, mp, DL_OUTSTATE) == RETRY)
			return(RETRY);
		return(DONE);
	}
	reqp = (dl_bind_req_t *)mp->b_rptr;
	/* don't check dl_service_mode -- IP doesn't set it */
	for (tslp = sliplist; tslp != NULL; tslp = tslp->s_next) {
		if (tslp->s_sap == reqp->dl_sap) {
			if (slip_errorack(slp, mp, DL_NOADDR) == RETRY)
				return(RETRY);
			return(DONE);
		}
	}
	bp = allocb(DL_BIND_ACK_SIZE+SLIPSAPSZ, BPRI_HI);
	if (bp == NULL) {
		slip_wsched(slp, DL_BIND_ACK_SIZE+SLIPSAPSZ);
		return(RETRY);
	}
	slp->s_sap = reqp->dl_sap;
	freemsg(mp);
	bp->b_datap->db_type = M_PCPROTO;
	ackp = (dl_bind_ack_t *)bp->b_wptr;
	ackp->dl_primitive = DL_BIND_ACK;
	ackp->dl_sap = slp->s_sap;
	ackp->dl_addr_length = SLIPSAPSZ;
	ackp->dl_addr_offset = DL_BIND_ACK_SIZE;
	ackp->dl_max_conind = 0;
#ifdef DLPIV2
	ackp->dl_xidtest_flg = 0;
#else
	ackp->dl_growth = 0;
#endif
	bp->b_wptr += DL_BIND_ACK_SIZE;
	*(ulong_t *)bp->b_wptr = slp->s_sap;
	bp->b_wptr += SLIPSAPSZ;
	slp->s_dlstate = DL_IDLE;
	putnext(slp->s_rdq, bp);
	return(DONE);
}

STATIC retval_t
slip_unbind(slip_t *slp, mblk_t *mp)
{
	mblk_t *bp;
	dl_ok_ack_t *ackp;

	if (slp->s_dlstate != DL_IDLE) {
		if (slip_errorack(slp, mp, DL_OUTSTATE) == RETRY)
			return(RETRY);
		return(DONE);
	}
	if ((bp = allocb(DL_OK_ACK_SIZE, BPRI_HI)) == NULL) {
		slip_wsched(slp, DL_OK_ACK_SIZE);
		return(RETRY);
	}
	freemsg(mp);
	putctl1(slp->s_rdq->q_next, M_FLUSH, FLUSHRW);
	bp->b_datap->db_type = M_PCPROTO;
	ackp = (dl_ok_ack_t *)bp->b_wptr;
	ackp->dl_primitive = DL_OK_ACK;
	ackp->dl_correct_primitive = DL_UNBIND_REQ;
	bp->b_wptr += DL_OK_ACK_SIZE;
	slp->s_dlstate = DL_UNBOUND;
	slp->s_sap = 0;
	putnext(slp->s_rdq, bp);
	return(DONE);
}

STATIC retval_t
slip_send(slip_t *slp, mblk_t *mp)
{
	mblk_t *bp, *omp;
	int size;

	if (slp->s_dlstate != DL_IDLE) {
		freemsg(mp);
		return(DONE);
	}
	if (!canput(WR(slp->s_rdq)->q_next))
		return(RETRY);
	size = msgdsize(mp);
	if ((size < MINSDUSZ) || (size > MAXSDUSZ)) {
		freemsg(mp);
		slp->s_ifstats.ifs_oerrors++;
		return(DONE);
	}
	size = 2*size + 2;	/* worst case */
	bp = allocb(size, BPRI_MED);
	if (bp == NULL) {
		slip_wsched(slp, size);
		return(RETRY);
	}
	omp = mp;
	mp = mp->b_cont;
	freeb(omp);
	omp = mp;
	*bp->b_wptr++ = END;
	while (mp != NULL) {
		while (mp->b_rptr < mp->b_wptr) {
			if (*mp->b_rptr == END) {
				*bp->b_wptr++ = ESC;
				*bp->b_wptr++ = ESC_END;
			} else if (*mp->b_rptr == ESC) {
				*bp->b_wptr++ = ESC;
				*bp->b_wptr++ = ESC_ESC;
			} else {
				*bp->b_wptr++ = *mp->b_rptr;
			}
			mp->b_rptr++;
		}
		mp = mp->b_cont;
	}
	*bp->b_wptr++ = END;
	freemsg(omp);
	putnext(WR(slp->s_rdq), bp);
	slp->s_ifstats.ifs_opackets++;
	return(DONE);
}

STATIC retval_t
slip_errorack(slip_t *slp, mblk_t *mp, ulong_t err)
{
	union DL_primitives *dp;
	dl_error_ack_t *errp;
	mblk_t *bp;

	bp = allocb(DL_ERROR_ACK_SIZE, BPRI_HI);
	if (bp == NULL) {
		slip_wsched(slp, DL_ERROR_ACK_SIZE);
		return(RETRY);
	}
	bp->b_datap->db_type = M_PCPROTO;
	errp = (dl_error_ack_t *)bp->b_wptr;
	errp->dl_primitive = DL_ERROR_ACK;
	dp = (union DL_primitives *)mp->b_rptr;
	errp->dl_error_primitive = dp->dl_primitive;
	errp->dl_errno = err;
	errp->dl_unix_errno = 0;
	bp->b_wptr += DL_ERROR_ACK_SIZE;
	freemsg(mp);
	putnext(slp->s_rdq, bp);
	return(DONE);
}

STATIC void
slip_hangup(queue_t *q)
{
	slip_t *slp;
	mblk_t *bp;

	slp = (slip_t *)q->q_ptr;
	if (slp->s_udmp != NULL) {
		freemsg(slp->s_udmp);
		slp->s_udmp = NULL;
	}
	if ((bp = slp->s_hupmp) == NULL)
		return;
	slp->s_hupmp = NULL;
	if (slipadmin.s_rdq == NULL) {
		freeb(bp);
	} else {
		*(dev_t *)bp->b_wptr = slp->s_dev;
		bp->b_wptr += sizeof(dev_t);
		putnext(slipadmin.s_rdq, bp);
		flushq(slp->s_rdq, FLUSHALL);
		flushq(WR(slp->s_rdq), FLUSHALL);
	}
}

STATIC int
sliprput(queue_t *q, mblk_t *mp)
{
	switch (mp->b_datap->db_type) {
	case M_FLUSH:
		if (*mp->b_rptr & FLUSHR)
			flushq(q, FLUSHALL);
		if (*mp->b_rptr & FLUSHW)
			flushq(WR(q), FLUSHALL);
		putnext(q, mp);
		break;

	case M_ERROR:
	case M_HANGUP:
	case M_DATA:
		putq(q, mp);
		break;

	case M_CTL:
		freemsg(mp);
		break;

	default:
		putnext(q, mp);
		break;
	}
	return(0);
}

STATIC int
sliprsrv(queue_t *q)
{
	mblk_t *mp, *udbp, *bp, *omp;
	slip_t *slp;
	dl_unitdata_ind_t *dp;

	slp = (slip_t *)q->q_ptr;
	while ((mp = getq(q)) != NULL) {
		if ((mp->b_datap->db_type == M_HANGUP) ||
		  (mp->b_datap->db_type == M_ERROR)) {
			flushq(q, FLUSHALL);
			slip_hangup(q);
			putnext(q, mp);
			return(0);
		}
		if (!canput(q->q_next)) {
			putbq(q, mp);
			return(0);
		}
		if (slp->s_udmp == NULL) {
			udbp = allocb(DL_UNITDATA_IND_SIZE, BPRI_MED);
			if (udbp == NULL) {
				slip_rsched(slp, DL_UNITDATA_IND_SIZE);
				putbq(q, mp);
				return(0);
			}
			bp = allocb(MAXSDUSZ, BPRI_MED);
			if (bp == NULL) {
				freeb(udbp);
				slip_rsched(slp, 0);
				putbq(q, mp);
				return(0);
			}
			udbp->b_datap->db_type = M_PROTO;
			dp = (dl_unitdata_ind_t *)udbp->b_wptr;
			dp->dl_primitive = DL_UNITDATA_IND;
			dp->dl_dest_addr_length = 0;
			dp->dl_dest_addr_offset = 0;
			dp->dl_src_addr_length = 0;
			dp->dl_src_addr_offset = 0;
#ifdef DLPIV2
			dp->dl_group_address = 0;
#else
			dp->dl_reserved = 0;
#endif
			udbp->b_wptr += DL_UNITDATA_IND_SIZE;
			udbp->b_cont = bp;
			slp->s_udmp = udbp;
		}
		omp = mp;
		udbp = slp->s_udmp->b_cont;
		while (mp != NULL) {
			while (mp->b_rptr < mp->b_wptr) {
				if (slp->s_flag & GOTESC) {
					if (*mp->b_rptr == ESC_END)
						*mp->b_rptr = END;
					else if (*mp->b_rptr == ESC_ESC)
						*mp->b_rptr = ESC;
					else
						slp->s_ifstats.ifs_ierrors++;
					if ((udbp->b_wptr - udbp->b_rptr) ==
					  MAXSDUSZ)
						mp->b_rptr++;
					else
						*udbp->b_wptr++ = *mp->b_rptr++;
					slp->s_flag &= ~GOTESC;
				} else if (*mp->b_rptr == END) {
					mp->b_rptr++;
					if (udbp->b_rptr != udbp->b_wptr) {
						putnext(q, slp->s_udmp);
						slp->s_ifstats.ifs_ipackets++;
						slp->s_udmp = NULL;
						if (msgdsize(omp) != 0)
							putbq(q, omp);
						else
							freemsg(omp);
						omp = mp = NULL;
						break;
					}
				} else if (*mp->b_rptr == ESC) {
					slp->s_flag |= GOTESC;
					mp->b_rptr++;
				} else {
					if ((udbp->b_wptr - udbp->b_rptr) ==
					  MAXSDUSZ) {
						mp->b_rptr++;
						slp->s_ifstats.ifs_ierrors++;
					} else {
						*udbp->b_wptr++ = *mp->b_rptr++;
					}
				}
			}
			if (mp != NULL)
				mp = mp->b_cont;
		}
		if (omp != NULL)
			freemsg(omp);
	}
	return(0);
}

STATIC void
slip_rsched(slip_t *slp, int size)
{
	ulong_t id;

	if (slp->s_flag & (RDBCALL|RDTOUT))
		return;
	if (size == 0)
		id = 0;
	else
		id = bufcall(size, BPRI_HI, slip_rcont, slp);
	if (id == 0) {
		id = timeout(slip_rcont, (caddr_t)slp,
		  drv_usectohz(HALFSEC));
		slp->s_flag |= RDTOUT;
	} else {
		slp->s_flag |= RDBCALL;
	}
	slp->s_rid = id;
}

STATIC void
slip_rcont(slip_t *slp)
{
	qenable(slp->s_rdq);
	slp->s_flag &= ~(RDBCALL|RDTOUT);
}
